part of pegdart;

abstract class Ignore {
  const Ignore();
}

class DelayedRangeGetter {

  final int lower, upper;
  final dynamic indexable;

  const DelayedRangeGetter(this.lower, this.upper, this.indexable);

  dynamic get match {
    if (indexable is String) {
      return indexable.substring(lower, upper);
    }
    return indexable.getRange(lower, upper).toList();
  }
}

class SeqMatcher extends PEGExpr {

  final dynamic seq;
  SeqMatcher(this.seq);

  dynamic matcherAt(int index) {
    try {
      return seq[index];
    } catch (exception) {
      return null;
    }
  }

  @override
  DelayedRangeGetter parse(_, InputContext ctx) {
    final int startPosition = ctx.currentPosition;
    int index = -1;
    dynamic currentChar = ctx.currentChar,
        char;

    while ((char = matcherAt(++index)) != null) {
      if (char == currentChar) {
        currentChar = ctx.advancePosition();
      } else {
        return null;
      }
    }

    return new DelayedRangeGetter(startPosition, ctx.currentPosition, ctx.indexable);
  }
  
}

class RegexMatcher extends PEGExpr {
  
  RegExp matcher;
  
  RegexMatcher(String regex) {
    matcher = new RegExp(regex);
  }
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final dynamic char = ctx.currentChar;
    try {
      // char can be null so we need to basically catch and ignore that case by returning null
      if (matcher.hasMatch(char)) {
        ctx.advancePosition();
        return char;
      }
    } catch (exception) {
      
    }
    return null;
  }
  
}

class SeqExpression extends PEGExpr {

  final List<PEGExpr> sequence = [];

  SeqExpression(PEGExpr left, PEGExpr right) {
    sequence..add(left)..add(right);
  }

  @override
  SeqExpression operator +(PEGExpr right) {
    sequence.add(right);
    return this;
  }

  PEGExpr parserAt(int index) {
    try {
      return sequence[index];
    } catch (exception) {
      return null;
    }
  }

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final List<dynamic> resultAccumulator = [];
    final Scope scopeAccumulator = scope.increment();
    int index = -1;
    PEGExpr seq;

    while ((seq = parserAt(++index)) != null) {
      final dynamic result = seq.parse(scopeAccumulator, ctx);
      if (result == null) {
        return null;
      }
      if (result != Ignore) {
        resultAccumulator.add(result);
      }
    }

    scope.merge(scopeAccumulator);
    return resultAccumulator;
  }
  
}

class AltExpression extends PEGExpr {

  final List<dynamic> alternatives = [];

  AltExpression(PEGExpr left, PEGExpr right) {
    alternatives..add(left)..add(right);
  }

  @override
  PEGExpr operator |(PEGExpr another) {
    alternatives.add(another);
    return this;
  }

  PEGExpr parserAt(int index) {
    try {
      return alternatives[index];
    } catch (exception) {
      return null;
    }
  }

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final int startPosition = ctx.currentPosition;
    int index = -1;
    PEGExpr currentParser;

    while ((currentParser = parserAt(++index)) != null) {
      final Scope altScope = scope.increment();
      final dynamic result = currentParser.parse(altScope, ctx);
      if (result == null) {
        ctx.jumpTo(startPosition);
      } else {
        scope.merge(altScope);
        return result;
      }
    }

    return null;
  }

}

class NegativeLookaheadExpression extends PEGExpr {

  final PEGExpr expression;

  NegativeLookaheadExpression(this.expression);

  @override
  PEGExpr operator -() => throw new Error();

  @override
  PEGExpr operator ~() => throw new Error();

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final int startPosition = ctx.currentPosition;
    final dynamic result = expression.parse(scope, ctx) == null ? Ignore : null;
    ctx.jumpTo(startPosition);
    return result;
  }

}

class PositiveLookaheadExpression extends PEGExpr {

  final PEGExpr expression;

  PositiveLookaheadExpression(this.expression);

  @override
  PEGExpr operator -() => throw new Error();

  @override
  PEGExpr operator ~() => throw new Error();

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final int startPosition = ctx.currentPosition;
    final dynamic result = expression.parse(scope, ctx) != null ? Ignore : null;
    ctx.jumpTo(startPosition);
    return result;
  }

}

class AnyExpression extends PEGExpr {

  final PEGExpr expression;

  AnyExpression(this.expression);

  @override
  PEGExpr get any => throw new Error();

  @override
  PEGExpr get many => throw new Error();

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final int startPosition = ctx.currentPosition;
    final dynamic result = expression.parse(scope, ctx);
    if (result == null) {
      ctx.jumpTo(startPosition);
      return Ignore;
    }
    return result;
  }

}

class ManyExpression extends PEGExpr {

  final PEGExpr expression;

  ManyExpression(this.expression);

  @override
  PEGExpr get many => throw new Error();

  Type parseIgnoring(Scope scope, InputContext ctx) {
    int startPosition = ctx.currentPosition;
    while (expression.parse(scope, ctx) != null) {
      /* This means we made no progress so if we continue we are going to be stuck in a loop. */
      if (ctx.currentPosition == startPosition) {
        throw new Error();
      }
      startPosition = ctx.currentPosition;
    }
    ctx.jumpTo(startPosition);
    return Ignore;
  }

  List<dynamic> parseAccumulating(Scope scope, InputContext ctx, List<dynamic> accumulator) {
    while (true) {
      final int startPosition = ctx.currentPosition;
      final dynamic result = expression.parse(scope, ctx);
      if (result == null) {
        ctx.jumpTo(startPosition);
        break;
      }
      /* Similar to above. This usually indicates some kind of bug in the grammar. */
      if (ctx.currentPosition == startPosition) {
        throw new Error();
      }
      accumulator.add(result);
    }
    return accumulator;
  }

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final List<dynamic> accumulator = [];
    dynamic result = expression.parse(scope, ctx);
    if (result == null) {
      return null;
    }
    if (result == Ignore) {
      return parseIgnoring(scope, ctx);
    }
    accumulator.add(result);
    return parseAccumulating(scope, ctx, accumulator);
  }

}

class IgnoreExpression extends PEGExpr {

  final PEGExpr expression;

  IgnoreExpression(this.expression);

  @override
  PEGExpr get ignore => throw new Error();

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    return expression.parse(scope, ctx) == null ? null : Ignore;
  }

}

class TracingExpression extends PEGExpr {

  final PEGExpr expression;
  final Symbol name;
  
  TracingExpression(this.expression, this.name);

  void traceCtxData(String message, InputContext ctx) {
    [message, "Position = ${ctx.currentPosition}", 
     "Character = ${ctx.currentChar == null ? '\$' : ctx.currentChar}", 
     "Cut! = ${ctx.cutPosition}"].forEach((m) => print("${name}: ${m}"));
  }

  @override
  dynamic parse(Scope scope, InputContext ctx) {
    print('');
    traceCtxData('Before', ctx);
    final dynamic result = expression.parse(scope, ctx);
    print('-------------');
    traceCtxData('After', ctx);
    print('');
    return result;
  }

}

class NamedExpression extends PEGExpr {
  
  final PEGExpr expression;
  final Symbol name;
  
  NamedExpression(this.expression, this.name);
  
  @override
  PEGExpr operator [](Symbol name) => throw new Error();
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final dynamic result = expression.parse(scope, ctx);
    scope[name] = result;
    return result;
  }
  
}

typedef dynamic PEGAction(Scope scope);
class ActionExpression extends PEGExpr {
  
  final PEGExpr expression;
  final PEGAction code;
  
  ActionExpression(this.expression, this.code);
  
  @override
  PEGExpr operator >>(PEGAction code) => throw new Error();
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final Scope actionScope = scope.increment();
    final dynamic result = expression.parse(actionScope, ctx);
    if (result == null) {
      return null;
    } else {
      final dynamic actionResult = code(actionScope);
      if (expression.name != null) {
        scope[expression.name] = actionResult; 
      }
      return actionResult;
    }
  }
  
}

typedef dynamic PEGErrorAction(Scope scope, InputContext ctx, Exception e);
class ErrorActionExpression extends PEGExpr {
  
  final PEGExpr expression;
  final PEGErrorAction code;
  
  @override
  PEGExpr operator >>(PEGAction code) => throw new Error();
  
  @override
  PEGExpr operator <<(PEGErrorAction code) => throw new Error();
  
  ErrorActionExpression(this.expression, this.code);
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    try {
      final dynamic result = expression.parse(scope, ctx);
      if (result == null) {
        final dynamic recoveryResult = code(scope, ctx, null);
        return recoveryResult;
      } else {
        return result;
      }
    } catch (exception) {
      code(scope, ctx, exception); throw exception;
    }
  }
}

typedef PEGExpr ParserCallback(Grammar g, Scope s, InputContext ctx);
class RuleExpression extends PEGExpr {
  
  final Symbol name;
  
  RuleExpression(this.name);
  
  @override
  dynamic parse(Scope scope, InputContext ctx) => ctx.rules[name].parse(scope, ctx);
  
}

class DelayExpression extends PEGExpr {
  
  final ParserCallback parserCreator;
  final Grammar grammar;
  
  DelayExpression(this.grammar, this.parserCreator);
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final PEGExpr parserExpression = parserCreator(grammar, scope, ctx);
    return parserExpression.parse(scope, ctx);
  }
  
}

class WildcardExpression extends PEGExpr {
  
  WildcardExpression();
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final dynamic currentChar = ctx.currentChar;
    if (currentChar == null) {
      return null;
    }
    ctx.advancePosition();
    return currentChar;
  }
  
}

class CutExpression extends PEGExpr {

  @override
  ManyExpression get many => throw new Error();
  
  @override
  AnyExpression get any => throw new Error();
  
  @override
  NamedExpression operator [](Symbol s) => throw new Error();
  
  @override
  ActionExpression operator >>(PEGAction c) => throw new Error();
  
  @override
  ErrorActionExpression operator <<(PEGErrorAction c) => throw new Error();
  
  @override
  PositiveLookaheadExpression operator ~() => throw new Error();
  
  @override
  NegativeLookaheadExpression operator -() => throw new Error();
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    ctx.cut();
    return Ignore;
  }
  
}

class RangeExpression extends PEGExpr {
  
  final List<PEGExpr> matchers;
  
  RangeExpression(this.matchers);
  
  @override
  dynamic parse(Scope scope, InputContext ctx) {
    final int startPosition = ctx.currentPosition;
    for (PEGExpr matcher in matchers) {
      final dynamic result = matcher.parse(scope, ctx);
      if (result != null) {
        return result;
      }
      ctx.jumpTo(startPosition);
    }
    return null;
  }
  
}

abstract class PEGExpr {
  
  Symbol name;

  SeqExpression operator +(PEGExpr other) => new SeqExpression(this, other);

  AltExpression operator |(PEGExpr other) => new AltExpression(this, other);

  NegativeLookaheadExpression operator -() => new NegativeLookaheadExpression(this);

  PositiveLookaheadExpression operator ~() => new PositiveLookaheadExpression(this);

  AnyExpression get any => new AnyExpression(this);

  ManyExpression get many => new ManyExpression(this);

  IgnoreExpression get ignore => new IgnoreExpression(this);

  TracingExpression trace(Symbol name) => new TracingExpression(this, name);

  NamedExpression operator [](Symbol name) => new NamedExpression(this, name);

  ActionExpression operator >>(PEGAction code) => new ActionExpression(this, code);

  ErrorActionExpression operator <<(PEGErrorAction code) => new ErrorActionExpression(this, code);

  dynamic parse(Scope scope, InputContext ctx);

}